## See https://packaging.python.org/guides/single-sourcing-package-version/

'''A collection of functions for ipython use in the classroom

   -------   Time-stamp: <2021-07-02T17:50:29.307889-04:00 cws2>   ------  '''


__version__ = "0.0b2.dev32" # X.YaN.devM format for now

